
#import "AliRtcPlayerExtension.h"
#import <AliyunPlayer/CicadaRenderDelegate.h>
#import <AliyunPlayer/CicadaRenderingDelegate.h>
#import <AliyunPlayer/CicadaAudioSessionDelegate.h>
#import <AliyunPlayer/CicadaFrameInfo.h>
#import <AliRTCSdk/AliRtcLog.h>
#import <atomic>

@interface AliRtcPlayerExtension() <CicadaRenderingDelegate, CicadaAudioSessionDelegate, AVPEventReportParamsDelegate>
{
    std::atomic<bool> isPublishAudio;
    std::atomic<bool> isPublishVideo;
    AliRtcVideoSource videoSourceType;
    int32_t playoutVolume;
    int32_t publishVolume;
    bool firstAudioSend;
    bool firstVideoSend;
}

@property (nonatomic, weak)AliPlayer *player;
@property (nonatomic, weak)AliRtcEngine *rtcEngine;
@end

@implementation AliRtcPlayerExtension

static AliRtcPlayerExtension *instance = nil;
static BOOL enableLog = NO;
static BOOL bindEvent = NO;
/**
 * @brief 初始化AliRtcPlayerExtension
 */
+ (instancetype)shareInstance {
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        if (instance == NULL) {
            instance = [[AliRtcPlayerExtension alloc] init];
        }
    });
    return instance;
}

typedef void(^playerLogBlock) (AVPLogLevel logLevel,NSString *strLog);
static playerLogBlock logBlock = ^(AVPLogLevel logLevel,NSString* strLog) {
    if (!enableLog) {
        return;
    }
    switch (logLevel) {
        case LOG_LEVEL_FATAL:
        case LOG_LEVEL_ERROR:
            AliEngineFatalLog(strLog);
            break;
        default:
            break;
    }
};

+ (void)enablePlayerLogger:(BOOL)enable {
    if (enableLog == enable) {
        return;
    }
    enableLog = enable;
    if (enable) {
        [AliPlayer setLogCallbackInfo:LOG_LEVEL_ERROR callbackBlock:logBlock];
    }
}

- (void)onEventReportParams:(NSDictionary<NSString *, NSString *>*)params
{
    static bool firstPrint = true;
    if (firstPrint) {
        AliEngineDebugLog(@"onEventReportParams:%@", params);
        firstPrint = false;
    } else {
        AliEngineDebugLog(@"onEventReportParams:%@", [params objectForKey:@"args"]);
    }
}

- (instancetype)init {
    if (self = [super init]) {
        AliEngineDebugLog(@"init AliRtcPlayerExtension");
        isPublishAudio = false;
        isPublishVideo = false;
        firstAudioSend = false;
        firstVideoSend = false;
        videoSourceType = AliRtcVideosourceScreenShareType;
        self.player = nil;
        self.rtcEngine = nil;
        [AliPlayer setAudioSessionDelegate:self];
        [AliPlayerGlobalSettings enableHWAduioTempo:NO];
        [AliPlayerGlobalSettings forceAudioRendingFormat:true fmt:@"s16" channels:1 sample_rate:48000];
    }

    return self;
}

- (void)dealloc {
    [AliPlayer setAudioSessionDelegate:nil];
}

- (int)bindPlayerToRtcEngine:(AliPlayer *)player engine:(AliRtcEngine *)engine {
    if ((self.player == player) && (self.rtcEngine == engine)) {
        return 0;
    }
    if ((nil == player) || (nil == engine)) {
        AliEngineFatalLog(@"bindPlayerToRtcEngine param error, player:%p, engine:%p", player, engine);
        return -1;
    }
    
    isPublishAudio = false;
    isPublishVideo = false;
    videoSourceType = AliRtcVideosourceScreenShareType;
    self.player = player;
    self.rtcEngine = engine;
    playoutVolume = 100;
    publishVolume = 100;

    [self.player setRenderingDelegate:self];
    if (enableLog) {
        [self.player setEventReportParamsDelegate:self];
        bindEvent = YES;
    }
    [self.player setMuted:YES];
    
    AliEngineDebugLog(@"bindPlayerToRtcEngine player:%p, engine:%p", player, engine);
    return 0;
}

- (int)unbindPlayerFromRtcEngine {
    if (self.player) {
        [self.player setRenderingDelegate:nil];
        if (bindEvent) {
            [self.player setEventReportParamsDelegate:nil];
            bindEvent = NO;
        }
    }

    [self unpublishAudio];
    [self unpublishVideo];
    self.player = nil;
    self.rtcEngine = nil;
    AliEngineDebugLog(@"unbindPlayerFromRtcEngine");
    return 0;
}

- (int)publishAudio:(BOOL)onlyLocalPlay {
    if (nil == self.rtcEngine) {
        AliEngineFatalLog(@"publishAudio param error");
        return -1;
    }
    
    firstAudioSend = false;
    isPublishAudio = true;
    AliRtcExtAudioSourceConfig *audioConfig = [[AliRtcExtAudioSourceConfig alloc] init];
    audioConfig.enable = YES;
    audioConfig.publish = !onlyLocalPlay;
    audioConfig.playout = YES;
    audioConfig.mixedWithMic = YES;
    [self.rtcEngine setExternalAudioSource:audioConfig];
    [self.rtcEngine setExternalAudioPublishVolume:publishVolume];
    [self.rtcEngine setExternalAudioPlayoutVolume:playoutVolume];

    AliEngineDebugLog(@"publishAudio:%d", onlyLocalPlay);
    return 0;
}

- (int)unpublishAudio {
    if (nil == self.rtcEngine) {
        AliEngineFatalLog(@"publishAudio param error");
        return -1;
    }
    
    isPublishAudio = false;
    AliRtcExtAudioSourceConfig *audioConfig = [[AliRtcExtAudioSourceConfig alloc] init];
    audioConfig.enable = NO;
    [self.rtcEngine setExternalAudioSource:audioConfig];
    
    AliEngineDebugLog(@"unpublishAudio");
    return 0;
}

- (int)publishVideo:(AliRtcVideoSource)type renderMode:(AliRtcRenderMode)renderMode {
    if (nil == self.rtcEngine) {
        AliEngineFatalLog(@"publishAudio param error");
        return -1;
    }
    
    firstVideoSend = false;
    isPublishVideo = true;
    videoSourceType = type;
    [self.rtcEngine setExternalVideoSource:YES useTexture:NO sourceType:type renderMode:renderMode];
    [self.rtcEngine publish:^(int errCode) {
    }];
    
    AliEngineDebugLog(@"publishVideo type:%d mode:%d", type, renderMode);
    return 0;
}

- (int)unpublishVideo {
    if (nil == self.rtcEngine) {
        AliEngineFatalLog(@"publishAudio param error");
        return -1;
    }
    
    isPublishVideo = false;
    [self.rtcEngine setExternalVideoSource:NO useTexture:NO sourceType:videoSourceType renderMode:AliRtcRenderModeAuto];
    [self.rtcEngine publish:^(int errCode) {
    }];
    
    AliEngineDebugLog(@"unpublishVideo");
    return 0;
}

- (int)setExternalAudioPublishVolume:(int)volume {
    if (nil == self.rtcEngine) {
        AliEngineFatalLog(@"publishAudio param error");
        return -1;
    }
    
    publishVolume = volume;
    return [self.rtcEngine setExternalAudioPublishVolume:volume];
}

- (int)setExternalAudioPlayoutVolume:(int)volume {
    if (nil == self.rtcEngine) {
        AliEngineFatalLog(@"publishAudio param error");
        return -1;
    }

    playoutVolume = volume;
    return [self.rtcEngine setExternalAudioPlayoutVolume:volume];
}

- (BOOL)onRenderingVideo:(CicadaFrameInfo*) frameInfo
{
    if (!isPublishVideo) {
        return NO;
    }

    switch (frameInfo.video_format) {
        case CICADA_PIX_FMT_APPLE_PIXEL_BUFFER:
        {
            if (!firstVideoSend) {
                firstVideoSend = true;
                AliEngineDebugLog(@"onRenderingVideo, pixelBuffer:%p rotate:%d, dar:%f", frameInfo.video_pixelBuffer, frameInfo.video_rotate, frameInfo.video_dar);
            }
            
            AliRtcVideoDataSample *dataSample = [[AliRtcVideoDataSample alloc] init];
            dataSample.pixelBuffer = frameInfo.video_pixelBuffer;
            dataSample.type = AliRtcBufferType_CVPixelBuffer;
            dataSample.rotation = frameInfo.video_rotate;
            dataSample.displayAspectRatio = frameInfo.video_dar;
            [self.rtcEngine pushExternalVideoFrame:dataSample sourceType:videoSourceType];
        }
            break;
        case CICADA_PIX_FMT_YUV420P:
        {
            if (!firstVideoSend) {
                firstVideoSend = true;
                AliEngineDebugLog(@"onRenderingVideo, y:%p, u:%p, v:%p, width:%d height:%d, rotate:%d, dar:%f",
                                  frameInfo.video_data_addr[0], frameInfo.video_data_addr[1], frameInfo.video_data_addr[2],
                                  frameInfo.video_width, frameInfo.video_height, frameInfo.video_rotate, frameInfo.video_dar);
            }
            
            if ((3 == frameInfo.video_data_lineNum) && frameInfo.video_data_addr) {
                AliRtcVideoDataSample *dataSample = [[AliRtcVideoDataSample alloc] init];
                dataSample.type = AliRtcBufferType_Raw_Data;
                dataSample.format = AliRtcVideoFormat_I420;
                dataSample.dataLength = 0;
                dataSample.dataPtr = 0;
                dataSample.dataYPtr = (long)(frameInfo.video_data_addr[0]);
                dataSample.dataUPtr = (long)(frameInfo.video_data_addr[1]);
                dataSample.dataVPtr = (long)(frameInfo.video_data_addr[2]);
                dataSample.strideY = frameInfo.video_data_lineSize[0];
                dataSample.strideU = frameInfo.video_data_lineSize[1];
                dataSample.strideV = frameInfo.video_data_lineSize[2];
                dataSample.width = frameInfo.video_width;
                dataSample.height = frameInfo.video_height;
                dataSample.rotation = frameInfo.video_rotate;
                dataSample.displayAspectRatio = frameInfo.video_dar;
                [self.rtcEngine pushExternalVideoFrame:dataSample sourceType:videoSourceType];
            }
        }
            break;
        default:
            AliEngineFatalLog(@"onRenderingVideo unspport %d", frameInfo.video_format);
            break;
    }

    return NO;
}

typedef std::numeric_limits<int16_t> limits_int16;
static inline int16_t FloatToS16(float v) {
  if (v > 0)
    return v >= 1 ? limits_int16::max()
    : static_cast<int16_t>(v * limits_int16::max() + 0.5f);
  return v <= -1 ? limits_int16::min()
    : static_cast<int16_t>(-v * limits_int16::min() - 0.5f);
}

- (BOOL)onRenderingAudio:(CicadaFrameInfo*) frameInfo
{
    if (!isPublishAudio) {
        return NO;
    }
    
    AliRtcExtAudioFrame *audioFrame = [[AliRtcExtAudioFrame alloc] init];
    audioFrame.sampleRate = frameInfo.audio_sample_rate;;
    audioFrame.channels = frameInfo.audio_channels;
    
    switch (frameInfo.audio_format)
    {
    case CICADA_SAMPLE_FMT_S16:
    {
        if ((0 < frameInfo.audio_data_lineNum) && frameInfo.audio_data) {
            if (!firstAudioSend) {
                firstAudioSend = true;
                AliEngineDebugLog(@"onRenderingAudio, sample_rate:%d channels:%d, format:%d",
                           frameInfo.audio_sample_rate, frameInfo.audio_channels, frameInfo.audio_format);
            }
            
            int32_t bufferSize = frameInfo.audio_nb_samples * frameInfo.audio_channels * sizeof(int16_t);
            audioFrame.dataPtr = (long) (frameInfo.audio_data[0]);
            audioFrame.numOfSamples = bufferSize;
            audioFrame.timeStamp = 0;
            [self.rtcEngine pushExternalAudioFrameRawData:audioFrame];
        }
        break;
    }
    case CICADA_SAMPLE_FMT_FLT:
    {
        if ((0 < frameInfo.audio_data_lineNum) && frameInfo.audio_data) {
            if (!firstAudioSend) {
                firstAudioSend = true;
                AliEngineDebugLog(@"onRenderingAudio, sample_rate:%d channels:%d, format:%d",
                           frameInfo.audio_sample_rate, frameInfo.audio_channels, frameInfo.audio_format);
            }
            
            int32_t length = frameInfo.audio_nb_samples * frameInfo.audio_channels;
            int32_t bufferSize = length * sizeof(int16_t);
            int16_t *renderBuffer = new int16_t[length];
            float *fBuffer = (float *)frameInfo.audio_data[0];
            for (size_t i = 0; i < frameInfo.audio_nb_samples * frameInfo.audio_channels; ++i)
            {
                renderBuffer[i] = FloatToS16(fBuffer[i]);
            }

            audioFrame.dataPtr = (long) renderBuffer;
            audioFrame.numOfSamples = bufferSize;
            audioFrame.timeStamp = 0;
            int ret = [self.rtcEngine pushExternalAudioFrameRawData:audioFrame];
            delete[] renderBuffer;
            if (ret < 0) {
    //            NSLog(@"pushExternalAudioFrameRawData failed");
            }
        }
        break;
    }
    case CICADA_SAMPLE_FMT_S16P:
    case CICADA_SAMPLE_FMT_FLTP:
    default:
        AliEngineFatalLog(@"onRenderingAudio unsupport format:%d", frameInfo.audio_format);
        break;
    }

    return NO;
}

- (BOOL)onRenderingFrame:(CicadaFrameInfo*) frameInfo
{
    if (Cicada_FrameType_Audio == frameInfo.frameType) {
        return [self onRenderingAudio:frameInfo];
    } else if (Cicada_FrameType_Video == frameInfo.frameType) {
        return [self onRenderingVideo:frameInfo];
    }
    
    return NO;
}

- (BOOL)setActive:(BOOL)active error:(NSError **)outError
{
    if (active) {
        return [[AVAudioSession sharedInstance] setActive:active error:outError];
    }
    return YES;
}

- (BOOL)setCategory:(NSString *)category withOptions:(AVAudioSessionCategoryOptions)options error:(NSError **)outError
{
    return YES;
}

- (BOOL)setCategory:(AVAudioSessionCategory)category mode:(AVAudioSessionMode)mode routeSharingPolicy:(AVAudioSessionRouteSharingPolicy)policy options:(AVAudioSessionCategoryOptions)options error:(NSError **)outError
{
    return YES;
}

@end
